; ------------------------------------------------------------------------------------
; Easy Crypted File
; MaxSize: ~8 Exabyte
; inkl. MemoryCache, Encryption, Callbacks
; Author: Alexander Aigner
; PB 4.4+ x86/x64 Ascii/Unicode
; ------------------------------------------------------------------------------------

; ToDo:
; 
; -

; EnableExplicit

; ------------------------------------------------------------------------------------
;- Internal
; ------------------------------------------------------------------------------------

Prototype ECF_FileAccess(ECF_File, AccessType, Size.q, curSize.q, CacheAccess, XtraData)

Enumeration
  #ECF_Magic = $30464345 ; ECF0
  #ECF_Access_Read = 0
  #ECF_Access_Write = 1
  #ECF_Access_ReBlock = 2
  
  #ECF_DefaultCacheSize = 512*1024*1024
  #ECF_DefaultBlockSize = 256*1024
EndEnumeration

Structure ECF_Header
  Magic.l
  HeaderSize.w
  BlockSize.q
  BlockCount.q
EndStructure

Structure ECF_FileInfo
  *FileAccessCB.ECF_FileAccess
  FileAccessCBXtraData.i
  FileCache.i
  MinFileCacheLifeTime.i
  FileHandle.i
  FileOffset.q
  FileName$
  UseCache.i
  DelayedWrite.a
  Key$  
EndStructure

Structure ECF_Data
  Header.ECF_Header
  FileInfo.ECF_FileInfo
EndStructure

Declare ECF_WriteBlock(*ECF.ECF_Data, Block, *BlockMem, FreeMem = #False)

XIncludeFile "FastMem.pbi"
XIncludeFile "Cache.pbi"

Macro FileExists(_File)
  (FileSize(_File)>-1)  
EndMacro

Macro _FileAccessCB(_AccessType, _Size, _curSize, _CacheAccess, _XtraData)
  If *ECF\FileInfo\FileAccessCB
    *ECF\FileInfo\FileAccessCB(*ECF, _AccessType, _Size, _curSize, _CacheAccess, _XtraData) 
  EndIf  
EndMacro

; RC4 Verschlsselung
Procedure _ECF_RC4Crypter(*Mem.Ascii, MemSize, *Key.Character)
  Protected i, j, x, *Sp.Long, *keyP.Character  
  Protected Dim S.l(255)
  Protected Dim K.l(255)
  
  If *Key\c
    *Sp = @S()
    *keyP = *Key
    
    ; Arrays Initialisieren
    For i = 0 To 255
      *Sp\l = i
      *Sp + SizeOf(Long)
      If *keyP\c = 0
        *keyP = *Key
      EndIf
      K(i) = *keyP\c
      *keyP + SizeOf(Character)
    Next
    
    ; S - Box fllen
    For i = 0 To 255
      j = (j + S(i) + K(i)) & 255
      Swap S(i), S(j)
    Next
    
    i = 0
    j = 0
    For x = 0 To MemSize-1
      i = (i + 1) & 255
      j = (j + S(i)) & 255
      Swap S(i), S(j)
      *Mem\a ! S((S(i) + (S(j) & 255)) & 255)
      *Mem + SizeOf(Ascii)
    Next
  EndIf
EndProcedure

; ------------------------------------------------------------------------------------
;- Public
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;  Datei ffnen/erstellen
;  File$: Quelldatei. Falls vorhanden wird versucht sie zu ffnen, ansonst erstellt
;  Key$: Das Passwort fr das File
;  BlockSize: Blockgre der Verschlsselung, Default 256 Kb
;  MaximumCacheSize: Maximale Gre des Zwischenspeichers, Default 256 MByte
;
;  Rckgabe: ECF - Handle, 0 wenn die Datei nicht erstellt werden konnte/keine ECF Datei ist/beschdigt ist. 
; ------------------------------------------------------------------------------------
Procedure ECF_OpenFile(File$, Key$ = "", BlockSize.q = #ECF_DefaultBlockSize, MaximumCacheSize = #ECF_DefaultCacheSize)
  Protected i, *ECF.ECF_Data = AllocateMemory(SizeOf(ECF_Data))
  
  With *ECF
    \FileInfo\UseCache = (MaximumCacheSize Or #Null)
    
    If Not FileExists(File$) ; Nicht da => erstellen
      
      \FileInfo\FileHandle = CreateFile(#PB_Any, File$)
      If \FileInfo\FileHandle ; konnte erstellt werden
        ; File Informationen schreiben
        \FileInfo\FileName$ = File$
        \FileInfo\Key$ = Key$
        
        ; Header erstellen
        \Header\Magic = #ECF_Magic
        \Header\HeaderSize = SizeOf(ECF_Header)
        
        ; Defaultwerte setzen
        \Header\BlockCount = #Null
        \Header\BlockSize = BlockSize
        
        If \FileInfo\UseCache
          \FileInfo\FileCache = Cache_New(MaximumCacheSize) ; Cache erstellen
          Cache_SetECF(\FileInfo\FileCache, *ECF)
        EndIf
        
        ; Header schreiben
        WriteData(\FileInfo\FileHandle, @\Header, SizeOf(ECF_Header))
      Else
        FreeMemory(*ECF)
        *ECF = #Null
      EndIf
      
    Else  ; Ist da => Laden
      \FileInfo\FileHandle = OpenFile(#PB_Any, File$)
      If \FileInfo\FileHandle ; konnte geladen werden
        ; FileInfos schreiben
        \FileInfo\FileName$ = File$
        \FileInfo\Key$ = Key$
        
        ; Header laden
        ReadData(\FileInfo\FileHandle, @\Header, SizeOf(ECF_Header))
        
        If \Header\Magic <> #ECF_Magic ; Magic Prfen
          ; Datei schlieen
          CloseFile(\FileInfo\FileHandle)
          FreeMemory(*ECF)
          *ECF = #Null
        Else
          ; Passen Headerangaben & Datei zusammen?
          If Not Lof(\FileInfo\FileHandle) = (\Header\HeaderSize+(\Header\BlockCount*\Header\BlockSize))
            ; Datei schlieen
            CloseFile(\FileInfo\FileHandle)
            FreeMemory(*ECF)
            *ECF = #Null
          Else
            If \FileInfo\UseCache
              \FileInfo\FileCache = Cache_New(MaximumCacheSize) ; Cache erstellen
              Cache_SetECF(\FileInfo\FileCache, *ECF)
            EndIf
          EndIf
        EndIf
        
      Else
        ; Speicher freigeben
        FreeMemory(*ECF)
        *ECF = #Null
      EndIf
    EndIf
    
    If *ECF ; experimentell
      FileBuffersSize(*ECF\FileInfo\FileHandle, *ECF\Header\BlockSize)  
    EndIf
  EndWith
 
  ProcedureReturn *ECF
EndProcedure

; ------------------------------------------------------------------------------------
;  Datei schlieen
; *ECF: ECF Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_CloseFile(*ECF.ECF_Data)
  With *ECF
    ; Datei schlieen, speicher freigeben
    If \FileInfo\UseCache
      Cache_Free(\FileInfo\FileCache)
    EndIf
    FlushFileBuffers(\FileInfo\FileHandle)
    CloseFile(\FileInfo\FileHandle)
    FreeMemory(*ECF)  
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Setzt den Schreib/Lese Zugriffs Callback
;  *ECF: ECF Handle
;  *CB: Pointer auf Callback
;  XtraData: Extradaten fr Callback
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_SetAccessCallback(*ECF.ECF_Data, *CB.ECF_FileAccess, XtraData = #Null)
  With *ECF
    \FileInfo\FileAccessCB = *CB
    \FileInfo\FileAccessCBXtraData = XtraData
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Aktiviert das verzgerte schreiben
;  *ECF: ECF Handle
;  Enable: #False disabled, sonst aktiviert
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_Transaction(*ECF.ECF_Data)
  With *ECF
    \FileInfo\DelayedWrite = #True
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Schreibt den Cache sicher in die Datei
;  *ECF: ECF Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_Commit(*ECF.ECF_Data)
  With *ECF
    If \FileInfo\UseCache
      If \FileInfo\DelayedWrite = #True
        Cache_Commit(\FileInfo\FileCache)
        \FileInfo\DelayedWrite = #False
      EndIf
    EndIf
    FlushFileBuffers(\FileInfo\FileHandle)
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Konfiguriert den Speicher - Cache
;  *ECF: ECF Handle
;  MaximumSize: Maximale Gre des Cache in Bytes, fr Unbegrenzt auf #NULL setzen
;  AgeTolerance: AltersToleranz bei der beim Speicherfreirumen noch nach schlechteren Daten gesucht wird
;  AccessTolerance: ZugriffssToleranz bei der beim Speicherfreirumen noch nach schlechteren Daten gesucht wird
;  MinimumEntryLifeTime: Minimale Lebenszeit eines Eintrages in Sekunden
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_ConfigureCache(*ECF.ECF_Data, MaximumSize = #ECF_DefaultCacheSize, AgeTolerance = #Null, AccessTolerance = #Null, MinimumEntryLifeTime = 30)
  With *ECF
    \FileInfo\MinFileCacheLifeTime = MinimumEntryLifeTime
    Cache_Config(\FileInfo\FileCache, MaximumSize, AgeTolerance, AccessTolerance)
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Setzt den Schreib/Lese Zeiger
;  *ECF: ECF Handle
;  Position: neue Position des Zeigers
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_Seek(*ECF.ECF_Data, Position.q)
  With *ECF
    \FileInfo\FileOffset = Position
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Ermittelt die aktuelle Position in der Datei
;  *ECF: ECF Handle
;
;  Rckgabe: Position
; ------------------------------------------------------------------------------------
Procedure.q ECF_Loc(*ECF.ECF_Data)
  With *ECF
    ProcedureReturn  \FileInfo\FileOffset
  EndWith
EndProcedure


; ------------------------------------------------------------------------------------
;  Ermittelt die aktuelle Datei - Gre
;  *ECF: ECF Handle
;
;  Rckgabe: Datei - Gre
; ------------------------------------------------------------------------------------
Procedure.q ECF_FileSize(*ECF.ECF_Data)
  With *ECF
    ProcedureReturn \Header\BlockCount*\Header\BlockSize
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Datei ab der aktuellen Zeigerpostion abschneiden
;  *ECF: ECF Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------

Procedure ECF_Truncate(*ECF.ECF_Data)
  Protected TruncatePosition.q
  With *ECF
     ; Nach welchem Block muss die Datei abgschnitten werden? Headergre + Ganze Blcke + Block in derm der Zeiger steht
     TruncatePosition = \Header\HeaderSize+((Int(\FileInfo\FileOffset/\Header\BlockSize))*\Header\BlockSize)+(((\FileInfo\FileOffset%\Header\BlockSize) Or #Null)*\Header\BlockSize)
     FileSeek(\FileInfo\FileHandle, TruncatePosition)
     TruncateFile(\FileInfo\FileHandle)
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Daten schreiben
;  *ECF: ECF Handle
;  *CData: zu schreibende Daten
;  DataSize: Gre der Daten
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure ECF_WriteData(*ECF.ECF_Data, *CData, DataSize.q)
  Protected fOffset.q, wSize.q, cStart.q, dStart.q, Seek.q, strongRandom, readFirst
  Protected *cRData, *RData, Blocks.q, sBlock.q, eBlock.q, cBlock, *Mem, i, oldCount.q
  
  With *ECF
    
    oldCount = \Header\BlockCount

    
    ; Blcke berechnen:
    sBlock = Int(\FileInfo\FileOffset/\Header\BlockSize)
    eBlock = Round((\FileInfo\FileOffset+DataSize)/\Header\BlockSize, #PB_Round_Up)
    Blocks = eBlock - sBlock
    
    ; Temporrblcke erstellen
    *cRData = AllocateMemory(\Header\BlockSize)
    
    fOffset = \FileInfo\FileOffset-sBlock*\Header\BlockSize
    
    For cBlock = 1 To Blocks ; Blcke schreiben
      
      Seek = \Header\HeaderSize+(sBlock+cBlock-1)*\Header\BlockSize
      
      ; Datei zu klein
      While (\Header\HeaderSize+\Header\BlockCount*\Header\BlockSize)<Seek
        ; Datei vergrern

        FileSeek(\FileInfo\FileHandle, Lof(\FileInfo\FileHandle))
        
        If Not *RData ; Zufallsdaten erstellen
          *RData = AllocateMemory(\Header\BlockSize)
          strongRandom = OpenCryptRandom()
        EndIf
      
        ; Zufallsvektor erstellen
        If strongRandom ; "starke" Zufallszahlen
          CryptRandomData(*RData, \Header\BlockSize) 
        Else
          For i = 0 To \Header\BlockSize-1
            PokeA(*RData+i, Random(255))  
          Next
        EndIf
      
        ; Blcke schreiben
        WriteData(\FileInfo\FileHandle, *RData, \Header\BlockSize)
      
        \Header\BlockCount+1
      Wend
      
      ; Zur Position springen
      FileSeek(\FileInfo\FileHandle, Seek)
      If Seek = (\Header\BlockCount*\Header\BlockSize+\Header\HeaderSize)   ; Wir sind am Ende => wir schreiben einen neuen Block :D
        \Header\BlockCount+1
      EndIf
        
      ; Anfangsposition/Blockgre berechnen
      ; Startposition/ Gre berechnen 
      If Blocks = 1 ; Wir schreiben nur in einem Block
        wSize = DataSize
        cStart = #Null
        dStart = fOffset
        readFirst = #True
      ElseIf cBlock = 1 ; Wir sind im 1. Block
        wSize = \Header\BlockSize-fOffset
        cStart = #Null
        dStart = fOffset
        readFirst = #True
      ElseIf cBlock = Blocks ; Wir sind im letzen Block
        wSize = (\FileInfo\FileOffset+DataSize)-((eBlock-1)*\Header\BlockSize)
        cStart = DataSize - wSize
        dStart = #Null
        readFirst = #True
      Else  ; Wir sind in einem Zwischenblock
        wSize = \Header\BlockSize
        cStart = (cBlock-1)*\Header\BlockSize-fOffSet
        dStart = #Null
        readFirst = #False
      EndIf
      
      If readFirst ; Nur Teilblock schreiben, => Block muss erst gelesen werden
        If Not \FileInfo\UseCache
          ReadData(\FileInfo\FileHandle, *cRData, \Header\BlockSize) ; Block laden
          ; Block entschlsseln
          _ECF_RC4Crypter(*cRData, \Header\BlockSize, @\FileInfo\Key$)
          *Mem = *cRData
        ElseIf Cache_isCached(\FileInfo\FileCache, sBlock+cBlock-1) ; Speicher aus Cache laden
          *Mem = Cache_GetMem(\FileInfo\FileCache)
        Else  
          ReadData(\FileInfo\FileHandle, *cRData, \Header\BlockSize) ; Block laden
          ; Block entschlsseln
          _ECF_RC4Crypter(*cRData, \Header\BlockSize, @\FileInfo\Key$)
          *Mem = *cRData
        EndIf
      Else
        *Mem = *cRData
      EndIf
      
      CopyMemory(*CData+cStart, *Mem+dStart, wSize)
      
      ; Callback aufrufen
      _FileAccessCB(#ECF_Access_Write, DataSize, cStart+wSize, #Null, \FileInfo\FileAccessCBXtraData) 
      
      If \FileInfo\UseCache ; Cacheeintrag anlegen
        If Not Cache_isCached(\FileInfo\FileCache, sBlock+cBlock-1)
          Cache_AddEntry(\FileInfo\FileCache, sBlock+cBlock-1, *Mem, \Header\BlockSize)
        EndIf  
        If \FileInfo\DelayedWrite ; Nur im Speicher ndern und bei gelegenheit Fluschen
          If Cache_isCached(\FileInfo\FileCache, sBlock+cBlock-1) ; UNBEDINGT Ntig falls das AddEntry() fehlschlgt, kann passieren.
            Cache_SetChangedFlag(\FileInfo\FileCache) ; Eintrag auf gendert setzen
            Continue
          EndIf
        EndIf
      EndIf
      
      If *Mem <> *cRData ; Cachespeicher darf nicht verschlsselt werden
        CopyMemory(*Mem, *cRData,  \Header\BlockSize)
        *Mem = *cRData
      EndIf
      
      ; Block verschlssel
      _ECF_RC4Crypter(*Mem, \Header\BlockSize, @\FileInfo\Key$)
      
      ; Block schreiben
      FileSeek(\FileInfo\FileHandle, Seek)
      WriteData(\FileInfo\FileHandle, *Mem, \Header\BlockSize)
      
    Next
    
    ; Aufrumen
    
    ; Zufallsdaten freigeben
    If *RData ; Zufallsdaten
      FreeMemory(*RData)
      
      If strongRandom ; Sicheren Generator beenden
        CloseCryptRandom()   
      EndIf
    EndIf
    
    If oldCount <> \Header\BlockCount ; Datei wurde vergrert => Header aktualisieren
      ; Header neu schreiben
    
      FileSeek(\FileInfo\FileHandle,  #Null)
      WriteData(\FileInfo\FileHandle, @\Header, \Header\HeaderSize)
    EndIf
    
    ; Temporrblcke freigeben
    FreeMemory(*cRData)
    
    ; Buffer sicher auf Disk schreiben
    FlushFileBuffers(\FileInfo\FileHandle)
    
    ; Fileposition ndern
    \FileInfo\FileOffset + DataSize
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  gezielt einen Block schreiben, fr das Verzgerte schreiben, nur aus Cache.
;  *ECF: ECF Handle
;  Block: BlockNummer
;  *BlockMem: Speicher des Blocks
;
;  Rckgabe: #Null
; ------------------------------------------------------------------------------------
Procedure ECF_WriteBlock(*ECF.ECF_Data, Block, *tBlockMem, FreeMem = #False)
  Protected Seek, i, strongRandom, *RData, *BlockMem
  
  With *ECF
    
    If FreeMem
      *BlockMem = *tBlockMem
    Else
      *BlockMem = AllocateMemory(\Header\BlockSize)
      CopyMemory(*tBlockMem, *BlockMem, \Header\BlockSize)
    EndIf
    
    ; Mssen wir die Datei erst vergrern?
    Seek = \Header\HeaderSize + Block*\Header\BlockSize
    While (\Header\HeaderSize+\Header\BlockCount*\Header\BlockSize)<Seek
      ; Datei vergrern

      FileSeek(\FileInfo\FileHandle, Lof(\FileInfo\FileHandle))
       
      If Not *RData ; Zufallsdaten erstellen
        *RData = AllocateMemory(\Header\BlockSize)
        strongRandom = OpenCryptRandom()
      EndIf
      
        ; Zufallsvektor erstellen
      If strongRandom ; "starke" Zufallszahlen
        CryptRandomData(*RData, \Header\BlockSize) 
      Else
        For i = 0 To \Header\BlockSize-1
          PokeA(*RData+i, Random(255))  
        Next
      EndIf
      
        ; Blcke schreiben
      WriteData(\FileInfo\FileHandle, *RData, \Header\BlockSize)
      
      \Header\BlockCount+1
    Wend
    
    If *RData
      FreeMemory(*RData)
    EndIf
    
    ; schreiben
    FileSeek(\FileInfo\FileHandle, Seek)
    _ECF_RC4Crypter(*BlockMem, \Header\BlockSize, @\FileInfo\Key$)
    WriteData(\FileInfo\FileHandle, *BlockMem, \Header\BlockSize)
    
    If Not FreeMem
      FreeMemory(*BlockMem)
    EndIf
    
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Daten lesen
;  *ECF: ECF Handle
;  *CData: Puffer in dem gelesen werden soll
;  DataSize: Gre der Daten
;
;  Rckgabe: True wenn gelesen werden konnte, False wenn nicht
; ------------------------------------------------------------------------------------
Procedure ECF_ReadData(*ECF.ECF_Data, *CData, DataSize.q)
  Protected fOffset.q, wSize.q, cStart.q, dStart.q, *Mem
  Protected *cRData, Blocks.q, sBlock.q, eBlock.q, cBlock, isCached
  
  With *ECF
    If (\Header\BlockCount*\Header\BlockSize) < (\FileInfo\FileOffset+DataSize) ; Daten sind nicht vorhanden
      ProcedureReturn #Null  
    EndIf
    
    ; Blcke berechnen:
    sBlock = Int(\FileInfo\FileOffset/\Header\BlockSize)
    eBlock = Round((\FileInfo\FileOffset+DataSize)/\Header\BlockSize, #PB_Round_Up)
    Blocks = eBlock - sBlock
    
    ; Temporrblcke erstellen
    *cRData = AllocateMemory(\Header\BlockSize)
    
    fOffset = \FileInfo\FileOffset-sBlock*\Header\BlockSize
      
    For cBlock = 1 To Blocks ; Block fr Block durchgehen
      
      If \FileInfo\UseCache

        isCached = Cache_isCached(\FileInfo\FileCache, sBlock+cBlock-1)
      EndIf
      
      If Not isCached
        FileSeek(\FileInfo\FileHandle, \Header\HeaderSize+(sBlock+cBlock-1)*\Header\BlockSize)
        ReadData(\FileInfo\FileHandle, *cRData, \Header\BlockSize) ; Block laden
      
        ; Block entschlsseln
        _ECF_RC4Crypter(*cRData, \Header\BlockSize, @\FileInfo\Key$)
      
        ; Cachen
        If \FileInfo\UseCache
          Cache_AddEntry(\FileInfo\FileCache, sBlock+cBlock-1, *cRData, \Header\BlockSize,  \FileInfo\MinFileCacheLifeTime)
        EndIf
        
        *Mem = *cRData
      Else
        *Mem = Cache_GetMem(\FileInfo\FileCache)
      EndIf
      
      ; Startposition/ Gre berechnen 
      If Blocks = 1 ; Wir schreiben nur in einem Block
        wSize = DataSize
        cStart = #Null
        dStart = fOffset
      ElseIf cBlock = 1 ; Wir sind im 1. Block
        wSize = \Header\BlockSize-fOffset
        cStart = #Null
        dStart = fOffset
      ElseIf cBlock = Blocks ; Wir sind im letzen Block
        wSize = (\FileInfo\FileOffset+DataSize)-((eBlock-1)*\Header\BlockSize)
        cStart = DataSize - wSize
        dStart = #Null 
      Else  ; wir sind in einem Zwischenblock
        wSize = \Header\BlockSize
        cStart = (cBlock-1)*\Header\BlockSize-fOffSet
        dStart = #Null
      EndIf
      
      CopyMemory(*Mem+dStart, *CData+cStart, wSize)
      
      ; Callback aufrufen
      _FileAccessCB(#ECF_Access_Read, DataSize, cStart+wSize, isCached, \FileInfo\FileAccessCBXtraData) 
    Next
    
    ; Temporrblcke freigeben
    FreeMemory(*cRData)
  
    ; Fileposition ndern
    \FileInfo\FileOffset + DataSize
  EndWith

  ProcedureReturn #True
EndProcedure

; ------------------------------------------------------------------------------------
;  Gibt Cache fr die Blocks frei die die Daten belegen
;  *ECF: ECF Handle
;  DataSize: Gre der Daten
;
; Rckgabe: #Null
; ------------------------------------------------------------------------------------
Procedure ECF_ClearCache(*ECF.ECF_Data, DataSize.q)
  Protected fOffset.q, wSize.q
  Protected Blocks.q, sBlock.q, eBlock.q, cBlock
  
  With *ECF
    If ((\Header\BlockCount*\Header\BlockSize) < (\FileInfo\FileOffset+DataSize)) Or (Not \FileInfo\UseCache) ; Daten sind nicht vorhanden
      ProcedureReturn #Null  
    EndIf
    
    ; Blcke berechnen:
    sBlock = Int(\FileInfo\FileOffset/\Header\BlockSize)
    eBlock = Round((\FileInfo\FileOffset+DataSize)/\Header\BlockSize, #PB_Round_Up)
    Blocks = eBlock - sBlock
    
    ; Temporrblcke erstellen
    
    fOffset = \FileInfo\FileOffset-sBlock*\Header\BlockSize
      
    For cBlock = 1 To Blocks ; Block fr Block durchgehen
      
      ; Startposition/ Gre berechnen 
      If Blocks = 1 ; Wir schreiben nur in einem Block
        wSize = DataSize
      ElseIf cBlock = 1 ; Wir sind im 1. Block
        wSize = \Header\BlockSize-fOffset
      ElseIf cBlock = Blocks ; Wir sind im letzen Block
        wSize = (\FileInfo\FileOffset+DataSize)-((eBlock-1)*\Header\BlockSize)
      Else  ; wir sind in einem Zwischenblock
        wSize = \Header\BlockSize
      EndIf
      
      If wSize = \Header\BlockSize ; nur ganz belegte Blcke lschen
        Cache_DeleteEntry(\FileInfo\FileCache, sBlock+cBlock-1)  
      EndIf
      
    Next
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Daten sicher lschen
;  *ECF: ECF Handle
;  DataSize: Gre der Daten
;
;  Rckgabe: True wenn gelscht werden konnte, False wenn nicht
; ------------------------------------------------------------------------------------
Procedure ECF_SafeDelete(*ECF.ECF_Data, DataSize.q)
  Protected *Mem, i
  
  With *ECF
    If (\Header\BlockCount*\Header\BlockSize) < (\FileInfo\FileOffset+DataSize) ; Daten sind nicht vorhanden
      ProcedureReturn #Null  
    EndIf
    
    *Mem = AllocateMemory(DataSize)
    If *Mem
      
      ; Zufallsdaten erstellen
      If OpenCryptRandom() ; "starke" Zufallszahlen
        CryptRandomData(*Mem, DataSize) 
        CloseCryptRandom()
      Else
        For i = 0 To DataSize-1
          PokeA(*Mem+i, Random(255))  
        Next
      EndIf
      
      ; Cache leeren
      ; ECF_ClearCache(*ECF, DataSize)
      
      ; Daten berschreiben
      FileSeek(\FileInfo\FileHandle, \Header\HeaderSize+\FileInfo\FileOffset)
      WriteData(\FileInfo\FileHandle, *Mem, DataSize)
      ; Sicher berschreiben
      FlushFileBuffers(\FileInfo\FileHandle)
      
      ; Zeiger weitersetzen
      \FileInfo\FileOffset + DataSize
      
      FreeMemory(*Mem)
      
      ProcedureReturn #True
    EndIf

  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Datei neu Blocken
;  *ECF: ECF Handle
;  NewBlockSize: Neue Blockgre
;  TempFileName$: Temporrfile zum Reblocken, wenn leergelassen im Temporrverzeichnis
;
;  Rckgabe: True wenn geReblockt werden konnte, sonst #Null
; ------------------------------------------------------------------------------------
Procedure ECF_ReBlock(*ECF.ECF_Data, NewBlockSize.q = #ECF_DefaultBlockSize, TempFileName$ = "")
  Protected *Mem, TempFile
  Protected *cRData, Blocks.q, cBlock, isCached.q, newBlockCount.q, cNewBlocks.q
  
  With *ECF
    If (Not \Header\BlockCount*\Header\BlockSize) And (NewBlockCount <> \Header\BlockSize)
      ProcedureReturn #Null  
    EndIf
    
    ; neue Blockanzahl ermitteln
    newBlockCount = (Int((\Header\BlockCount*\Header\BlockSize)/NewBlockSize))+(((\Header\BlockCount*\Header\BlockSize)%NewBlockSize) Or #Null)
    
    ; Tempfile erstellen
    If Not TempFileName$
      TempFileName$ = GetTemporaryDirectory()+GetFilePart(\FileInfo\FileName$)+"_rbl"+Str(Random(9999))
    EndIf
    
    ; Tempfile erstellen
    TempFile = ECF_OpenFile(TempFileName$, \FileInfo\Key$, NewBlockSize)
    
    If TempFile
       ; Temporrblcke erstellen
       *cRData = AllocateMemory(\Header\BlockSize)
      
      For cBlock = 1 To \Header\BlockCount ; Block fr Block durchgehen
       
        If \FileInfo\UseCache
          isCached = Cache_isCached(\FileInfo\FileCache, cBlock-1)
        EndIf
      
        If Not isCached
          FileSeek(\FileInfo\FileHandle, \Header\HeaderSize+(cBlock-1)*\Header\BlockSize)
          ReadData(\FileInfo\FileHandle, *cRData, \Header\BlockSize) ; Block laden
      
          ; Block entschlsseln
          _ECF_RC4Crypter(*cRData, \Header\BlockSize, @\FileInfo\Key$)
      
          *Mem = *cRData
        Else
          *Mem = Cache_GetMem(\FileInfo\FileCache)
        EndIf
      
        ; Neue Datei schreiben
        ECF_WriteData(TempFile, *Mem, \Header\BlockSize)
      
        ; Callback aufrufen
        _FileAccessCB(#ECF_Access_Read, \Header\BlockCount*\Header\BlockSize, cBlock*\Header\BlockSize, isCached, \FileInfo\FileAccessCBXtraData) 
      Next
    
      ; Temporrblcke freigeben
      FreeMemory(*cRData)
    
      ; Temporrdatei schlieen
      ECF_CloseFile(TempFile)
    
      ;Aktuelle Datei schlieen
      CloseFile(\FileInfo\FileHandle)
      ; Aktuelle Datei lschen, neue Datei -> alter umbennen und ffnen.
      DeleteFile(\FileInfo\FileName$)
      RenameFile(TempFileName$, \FileInfo\FileName$)
      \FileInfo\FileHandle = OpenFile(#PB_Any, \FileInfo\FileName$)
      
      ; Header anpassen
      \Header\BlockCount = newBlockCount
      \Header\BlockSize = NewBlockSize
    
      ; Cache fluschen, die Eintrge sind nicht mehr gltig
      If \FileInfo\UseCache
        Cache_Flush(\FileInfo\FileCache)
      EndIf
    Else
      ProcedureReturn #Null  
    EndIf
    
  EndWith

  ProcedureReturn #True
EndProcedure


; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 58
; Folding = ----
; EnableXP
; DisableDebugger