; ------------------------------------------------------------------------------------
; Free Space Controller
; Author: Alexander Aigner
; Verwaltet Speicherbereiche und fgt zusammenhngende Speicherbereiche automatisch zu greren zusammen
; ------------------------------------------------------------------------------------

XIncludeFile "LinkedList.pbi"

; EnableExplicit

#FSC_Magic = $88668866

; FSC Eintrag
Structure FSC_Entry
  Offset.q
  Size.q
EndStructure

; FSC Daten
Structure FSC_Data
  *LList.LL_Data
  TotalFreeSpace.q
  *Current.FSC_Entry
EndStructure

Procedure FSC_New() ; Neuen Controller erstellen
  Protected *FSC.FSC_Data = AllocateMemory(SizeOf(FSC_Data))  
  With *FSC
    \LList = LL_New()
  EndWith
  ProcedureReturn *FSC
EndProcedure

Procedure FSC_Free(*FSC.FSC_Data) ; Controller freigeben
  Protected *FEntry.FSC_Entry
  With *FSC
    
    ; Alle Eintrge freigeben
    LL_Reset(\LList)
    While LL_Next(\LList)
      *FEntry = LL_Get(\LList)
      FreeMemory(*FEntry)
      LL_Delete(\LList)
    Wend

    LL_Free(\LList)
    
    FreeMemory(*FSC)
  EndWith
EndProcedure

Procedure.q FSC_TotalFreeSpace(*FSC.FSC_Data) ; Controller freigeben
  Protected *FEntry.FSC_Entry
  With *FSC
    ProcedureReturn \TotalFreeSpace
  EndWith
EndProcedure

; Eintrag hinzufgen
Procedure FSC_Add(*FSC.FSC_Data, OffSet.q, Size.q, _NoSearch = #False)
  Protected *FEntry.FSC_Entry, *tFEntry.FSC_Entry, Change
  With *FSC
    *FEntry = AllocateMemory(SizeOf(FSC_Entry))
    
    
    ; Zusammenhngende Bereiche zusammenfassen
    If Not _NoSearch ; Nur wenn nicht gerade ein Element gefunden wurde suchen
      Repeat
        Change = #False
        
        LL_Reset(\LList) ; Alle durchgehen
        While LL_Next(\LList)
          *tFEntry = LL_Get(\LList)
      
          If OffSet+Size = *tFEntry\Offset ; Bereich stt vorne an
            Size = Size + *tFEntry\Size
            Change = #True
            LL_Delete(\LList)
          ElseIf *tFEntry\Offset + *tFEntry\Size = OffSet ; Bereich stt hinten an
            OffSet = *tFEntry\Offset
            Size = Size + *tFEntry\Size
            Change = #True
            LL_Delete(\LList)
          EndIf
        Wend
      Until Not Change ; Solange durchgehen bis nichts mehr zusammengefast werden konnte
      
    EndIf
    
    *FEntry\Offset = Offset
    *FEntry\Size = Size
    \TotalFreeSpace + Size
    LL_Add(\LList) : LL_Set(\LList, *FEntry)
    
  EndWith
EndProcedure

; Platz finden
Procedure FSC_Find(*FSC.FSC_Data, Size.q, *Found.Integer, _TakeFirst = #False)
  Protected OffSet.q, minSize.q, Entry
  Protected *FEntry.FSC_Entry
  
  *Found\i = #False ; Wir haben mal nix gefunden ...
  
  With *FSC
    LL_Reset(\LList) ; Alle durchgehen
    While LL_Next(\LList)
      *FEntry = LL_Get(\LList)
      If *FEntry\Size >= Size ; Eintrag passt :D
        *Found\i = #True
        
        If *FEntry\Size = Size ; OPTIMUM!
          Entry = LL_GetEntry(\LList)
          minSize = *FEntry\Size
          OffSet = *FEntry\Offset
          Break
        ElseIf minSize = #Null Or minSize > *FEntry\Size ; besseres Element
          Entry = LL_GetEntry(\LList)
          minSize = *FEntry\Size
          OffSet = *FEntry\Offset
          If _TakeFirst ; Erstbestes Element nehmen
            Break  
          EndIf
        EndIf
      EndIf
      
    Wend
    
    If *Found\i ; Was gefunden :D
      ; Eintrag lschen
      \TotalFreeSpace - minSize
      LL_SetEntry(\LList, Entry)
      LL_Delete(\LList)
      If minSize-Size ; Rest wieder auf den Stapel legen
        FSC_Add(*FSC, OffSet+Size, minSize-Size, #True)  
      EndIf  
    EndIf
    
  EndWith
  
  ProcedureReturn Offset
EndProcedure

; Eintrge in einen Speicherblock packen
Procedure FSC_SaveToMem(*FSC.FSC_Data, *MemSize.Quad)
  Protected *Mem, Pos.q
  With *FSC
    *MemSize\q = LL_Count(\LList)*SizeOf(FSC_Entry)
    If *MemSize\q
      *Mem = AllocateMemory(*MemSize\q)
      LL_Reset(\LList)
      While LL_Next(\LList)
        CopyMemory(LL_Get(\LList), *Mem+(Pos*SizeOf(FSC_Entry)), SizeOf(FSC_Entry))
        Pos + 1  
      Wend
    EndIf
  EndWith
  
  ProcedureReturn *Mem
EndProcedure

; Eintrge aus einem Speicherblock laden
Procedure FSC_LoadFromMem(*FSC.FSC_Data, *Mem, MemSize, FreeMem = #True)
  Protected FEntry.FSC_Entry, Pos, Cnt.q
  With *FSC
    
    Cnt = (MemSize/SizeOf(FSC_Entry))-1
    
    LL_Reset(\LList)
    While LL_Next(\LList)
      FreeMemory(LL_Get(\LList)) 
    Wend
    
    For Pos = 0 To Cnt
      CopyMemory(*Mem+(Pos*SizeOf(FSC_Entry)), @FEntry, SizeOf(FSC_Entry))
      FSC_Add(*FSC, FEntry\Offset, FEntry\Size, #True)
    Next
    
    If FreeMem
      FreeMemory(*Mem)  
    EndIf
    
  EndWith
  
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 6
; FirstLine = 4
; Folding = --
; EnableXP