; ------------------------------------------------------------------------------------
; Growable Memory
; (c) Alexander Aigner
; PB 4.3 + / x86 + x64 / Ascii + Unicode
; ------------------------------------------------------------------------------------

; EnableExplicit

Structure GMem_Data
  *Mem
  MemSize.i
  cPos.i
  BlockCount.i
  BlockSize.i
EndStructure

Procedure GMem_New(BlockSize = 512)
  Protected *GMem.GMem_Data = AllocateMemory(SizeOf(GMem_Data))
  *GMem\BlockSize = BlockSize
  ProcedureReturn *GMem
EndProcedure

Procedure GMem_Free(*GMem.GMem_Data)
  With *GMem
    If \Mem
      FreeMemory(\Mem)  
    EndIf
    FreeMemory(*GMem)
  EndWith
EndProcedure

Procedure GMem_SetMem(*GMem.GMem_Data, *Mem, MemSize)
  With *GMem
    If \Mem
      FreeMemory(\Mem)  
    EndIf
    \Mem = *Mem
     \MemSize = MemSize
     \cPos = #Null 
  EndWith
EndProcedure

Procedure GMem_WriteData(*GMem.GMem_Data, *Mem, MemSize)
  With *GMem
    While (MemSize + \cPos) > (\BlockCount*\BlockSize)
      \BlockCount+1
      \Mem = ReAllocateMemory(\Mem, \BlockCount*\BlockSize)
    Wend
    
    \Mem = ReAllocateMemory(\Mem, \MemSize + MemSize)
     CopyMemory(*Mem, \Mem + \cPos, MemSize)
     \MemSize + MemSize
     \cPos + MemSize 
  EndWith
EndProcedure

Procedure GMem_ReadData(*GMem.GMem_Data, *Mem, MemSize)
  With *GMem
     CopyMemory(\Mem + \cPos, *Mem,  MemSize)
     \cPos + MemSize 
  EndWith
EndProcedure

Procedure GMem_GetMem(*GMem.GMem_Data)
  With *GMem
    ProcedureReturn \Mem
  EndWith
EndProcedure

Procedure GMem_GetMemSize(*GMem.GMem_Data)
  With *GMem
    ProcedureReturn \MemSize
  EndWith
EndProcedure

Procedure GMem_Seek(*GMem.GMem_Data, Seek)
  With *GMem
    \cPos = Seek
  EndWith
EndProcedure



; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 4
; Folding = --
; EnableXP