; ------------------------------------------------------------------------------------
; Simple Container
; Based on Easy Cryted File
; Author: Alexander Aigner
; PB 4.4+ x86/x64 Ascii/Unicode
; ------------------------------------------------------------------------------------

; ### Feature - Liste ###

; Beliebig viele Eintrge mglich, ber StringID ansprechbar
; Beliebig viele Extra-Eintrge zu jedem Eintrag mglich, ber StingID ansprechbar, Quad und String
; Flexibles Lschen von Eintrgen, Shrinken von freiem Speicher mglich, wird sonst wiederverwendet.
; StringID jedes Eintrages nderbar
; Sicheres Lschen von Einrgen mglich, sichere Crypt-Funktionen untersttzt
; Kalibrierbares, selbstlernendes Cache - System fr bessere Performance bei sehr aktiven Containern
; Callbacksystem fr Events und Fortschritte
; Jeder Eintrag kann mit einem anderen Packer komprimiert werden
; Der gesamte Container kann mit einer Verschlsselung geschtzt werden
; Container ist blockbasiert umd nicht immer die ganze Datei entschlsseln zu mssen, nderung der Blckgre (reBlock) mglich
; Transaction/Commit System um nderungen nur im Speicher durchzufhren
; Loggingsystem, ntlich fr Debugzwecke/Konsolen
; Auf- und Abwrtskompatibilitt
; Untersttzte Packer: zLib, LZMA, bZip2, aPLib
; Untersttzte Verschlsselungen: RC4
; Untersttzung fr x86 (32 Bit) und x64 (64 Bit); Optimiert fr 64 Bit
; Ascii und Unicode Kompatibel

; ToDo: natives Splitten
; ToDo: (spter) Callbacks Asyncron ablaufen lassen, BackgroundThread.

; Changelog V 1.01:
;   - Logger eingebaut, ntzlich fr Debugzwecke
;   - Transaction/Commit eingebaut, bei richtiger Verwendung viel Performancegewinn
;   - LinkedLists auf PB - Listen umgestellt, per Macros gewrappt, leichter Performancegewinn
;   - Header- & Indexkompression
;   - BugFix: Header & Index werden nur mehr neu geschrieben wenn sich die Datei auch wirklich ndert

; ------------------------------------------------------------------------------------
;- Internal
; ------------------------------------------------------------------------------------

Prototype SC_EventCallback(*SC, EventType, Size.q, curSize.q, CacheAccess)

; Ab dieser Anzahl von Eintrgen wird die HashTable zum Suchen genutz (Defaultwert)
#SC_HT_Limit = 200
; Ab wie vielen Zugriffen landet ein Eintrag im unpacked - Cache?
#SC_UD_Limit = 2
; Version
#SC_Version = 100

; Includes
XIncludeFile "FastMem.pbi" ; Schnelle Memoryfunktionen, berschreibt CopyMemory()
XIncludeFile "LinkedList.pbi" ; Dynamische Linked List
XIncludeFile "ECF.pbi" ; EFC - Dateiverwaltung
XIncludeFile "FreeSpaceController.pbi" ; Speicherplatzverwaltung
XIncludeFile "packer.pbi" ; Universelle Kompressionsbibliothek
XIncludeFile "HashTable.pbi" ; Hashtable um die Zugriffszeiten bei vielen Eintrgen zu Optimieren
XIncludeFile "GrowableMemory.pbi" ; Wachsender Speicher
XIncludeFile "Logger.pbi" ; Logsystem

EnableExplicit

Structure SC_Entry_XtraData_Entry_Dynamics ; Dynamische Eintrge fr die XtraDaten eines Entrys
  *Name ; Name des Entrys
  Value.q ; Wert
EndStructure

Structure SC_Entry_XtraData_Entry ; XtraDaten eines Entrys
  Type.a ; Typ (String/Quad)
  NameSize.l ; Name
  StringSize.l ; Gre des Strings, falls Type = String
  Dyn.SC_Entry_XtraData_Entry_Dynamics ; Dynamische Daten
EndStructure

Structure SC_Entry_Dynamics
  *Name ; Name
  *XtraList.LL_Data ; Liste fr Extra - Eintrge
  tId.q ; Temporre Eintrags - Id
  tAccessCount.i ; Aktueller Zugriffszhler
EndStructure

; Structures
Structure SC_Entry
  OwnSize.w ; Gre der Struktur: um bei nderungen der Struktur abwrts-kompatibel zu bleiben.
  Position.q ; Position des Eintrags in der Datei
  Size.q ; Unkomprimierte gre des Eintrags
  PackedSize.q ; Gepackte Gre des Eintrags
  Compressed.a ; sind die Daten komprimiert
  NameSize.l ; Gre des Namens
  XtraDataCount.l ; Anzahl der Extra - Eintrge
  Packer.a ; Welcher Packer wurde verwendet?
  ; <- Neue Werte hier einfgen
  Dyn.SC_Entry_Dynamics ; MUSS!!!! Am Ende stehen! enthlt Dynamische Daten
EndStructure

Structure SC_Current ; Aktuelles Entry
  *Entry.SC_Entry
  *Mem
EndStructure

Structure SC_Data
  EnumLoadMem.i ; Speicher bei Enumeration laden?
  *Entrys.LL_Data ; Eintrge - Liste
  EntryHT.i ; Eintrge - Hashtable
  HTLimit.i ; Ab wie vielen Eintrgen soll die Hashtable verwendet werden?
  UDLimit.i ; Ab wie vielen Zugriffen landet ein Eintrag im Unpacked - Cache?
  FSC.i ; FreeSpaceController
  ECF.i ; ECF - Dateiverwaltung
  UDCache.i ; Cache fr entpackte Daten
  UseUDCache.b ; Wird der Cache verwendet?
  Logger.i ; Logger
  EnableDelayedWrite.a ; Verzgertes schreiben aktivieren
  ContainerChanged.a ; Container wurde verndert
  CPos.q ; aktuelle Position in der Datei
  Current.SC_Current ; Aktueller Eintrag
  TIdIndex.q ; Aktueller TId - Counter
  *Event.SC_EventCallback ; Event - Callback
EndStructure

Structure SC_Head
  Magic.l ; Magic (ist das eine ECF - Datei?)
  OffSet.q ; Offset des Indexes 
  EntryCount.q ; Anzahl Eintrge
  FSCSize.q ; FreeSpace Size
EndStructure

; Enumerations
Enumeration
  #SC_DefaultBlockSize = #ECF_DefaultBlockSize
  #SC_DefaultCacheSize = #ECF_DefaultCacheSize
  #SC_Magic = $23651290
EndEnumeration

Enumeration
  #SC_Event_Read = #ECF_Access_Read
  #SC_Event_Write = #ECF_Access_Write
  #SC_Event_ReBlock = #ECF_Access_ReBlock
  #SC_Event_CompressStart
  #SC_Event_CompressFinished
EndEnumeration

Enumeration
  #SC_zLib = #Packer_zLib
  #SC_LZMA = #Packer_LZMA
  #SC_bZip2 = #Packer_bZip2
  #SC_aPLib = #Packer_aPLib
  #SC_Uncompressed = #Packer_Uncompressed
EndEnumeration

; Enpackt den Eintrag aus *SCE nach \Current\Mem
Macro _SC_LoadEntry(_SC)
  ; Eintrag laden
  Protected *pMem, isUDCached
  
  
  If _SC#\UseUDCache ; Suchen wir mal im Cache ...
    If Cache_isCached(_SC#\UDCache, *SCE\Dyn\tId)
      ; Eintrag vorhanden, einfach aus dem Cache laden, spart Entschlsseln & Entpacken :D
      _SC#\Current\Mem = AllocateMemory(*SCE\Size)
      CopyMemory(Cache_GetMem(_SC#\UDCache), _SC#\Current\Mem, *SCE\Size)
      isUDCached = #True
    EndIf
  EndIf
  
  If Not isUDCached ; Wir sind nicht im Cache, normal laden
    *pMem = AllocateMemory(*SCE\PackedSize)
    ECF_Seek(_SC#\ECF, *SCE\Position)
    ECF_ReadData(_SC#\ECF, *pMem, *SCE\PackedSize)
        
    ; Eintrag entpacke wenn ntig
    If *SCE\Compressed
      _SC#\Current\Mem = Packer_Uncompress(*pMem, *SCE\PackedSize, *SCE\Size, *SCE\Packer)
      FreeMemory(*pMem)
    Else
      _SC#\Current\Mem = *pMem
    EndIf
    
    ; Daten landen im Unpacked - Cache
    If _SC#\UDLimit <= *SCE\Dyn\tAccessCount
      Cache_AddEntry(_SC#\UDCache, *SCE\Dyn\tId, _SC#\Current\Mem, *SCE\Size)
      ; Lese/Schreibe Cache Fluschen, nun unntig.
      ECF_Seek(_SC#\ECF, *SCE\Position)
      ECF_ClearCache(_SC#\ECF, *SCE\PackedSize)
    EndIf
  EndIf
EndMacro

; Sucht nach einem Eintrag und gibt den *SCE Eintrag darauf zurck falls gefunden
Procedure _SC_FindEntry(*SC.SC_Data, *Name, _EnableHTSearch = #True)
  Protected *SCE.SC_Entry
  
  With *SC
    If (LL_Count(\Entrys) >= \HTLimit) And _EnableHTSearch ; Hashtablesuche bei vielen Eintrgen verwenden, bessere Performance
      ProcedureReturn HT_GetValue(\EntryHT, PeekS(*Name))  
    Else ; Sonst alle Listenelemente durchgehen
      LL_Reset(\Entrys)
      While LL_Next(\Entrys)
        *SCE = LL_Get(\Entrys)
        If (CompareMemoryString(*SCE\Dyn\Name, *Name) = #PB_String_Equal) ; Treffer
          ProcedureReturn *SCE
        EndIf
      Wend
    EndIf
  EndWith
EndProcedure

; XtraData - Eintrag finden
Procedure _SC_FindXtraData(*SC.SC_Data, *Name)
  Protected *SCEXE.SC_Entry_XtraData_Entry
  
  ; Liste durchgehen
  With *SC
    If \Current\Entry And \Current\Entry\Dyn\XtraList
      LL_Reset(\Current\Entry\Dyn\XtraList)
      While LL_Next(\Current\Entry\Dyn\XtraList)
        *SCEXE = LL_Get(\Current\Entry\Dyn\XtraList)
        If (CompareMemoryString(*SCEXE\Dyn\Name, *Name) = #PB_String_Equal) ; Treffer
          ProcedureReturn *SCEXE
        EndIf
      Wend
    EndIf
  EndWith
EndProcedure

; Extradaten zum Eintrag speichern
Procedure _SC_SetXtraData(*SC.SC_Data, Name$, Type.a, XData.q)
  Protected *SCEXE.SC_Entry_XtraData_Entry, AddToList = #True
  
  With *SC
    If \Current\Entry
      If Not \Current\Entry\Dyn\XtraList ; Liste anlegen wenn ntig
        \Current\Entry\Dyn\XtraList = LL_New()
      EndIf
      
      *SCEXE = _SC_FindXtraData(*SC, @Name$)
      
      ; Eintrag finden oder Erstellen wenn nicht vorhanden
      If *SCEXE
        AddToList = #False
        If *SCEXE\Type = #PB_String
          FreeMemory(*SCEXE\Dyn\Value)
          *SCEXE\StringSize = #Null
        EndIf
      Else
        ; Eintrag erstellen und Namen schreiben
        *SCEXE = AllocateMemory(SizeOf(SC_Entry_XtraData_Entry))
        *SCEXE\NameSize = StringByteLength(Name$) + SizeOf(Character)
        *SCEXE\Dyn\Name = AllocateMemory(*SCEXE\NameSize)
        PokeS(*SCEXE\Dyn\Name, Name$)
      EndIf
      
      ; Inhalt schreiben
      *SCEXE\Type = Type
      Select *SCEXE\Type
        Case #PB_String
          *SCEXE\StringSize = MemoryStringLength(XData) + SizeOf(Character)
          *SCEXE\Dyn\Value = AllocateMemory(*SCEXE\StringSize)
          CopyMemory(XData, *SCEXE\Dyn\Value, *SCEXE\StringSize)
        Default
          *SCEXE\Dyn\Value = XData
      EndSelect
      
      If AddToList ; zu ExtraDaten hinzufgen
        LL_Add(\Current\Entry\Dyn\XtraList) : LL_Set(\Current\Entry\Dyn\XtraList, *SCEXE)
      EndIf
      
      ; Container wurde verndert
      \ContainerChanged = #True
      
      ProcedureReturn #True
    EndIf
  EndWith  
EndProcedure

; Setzt ein Event ab.
Macro _SC_SendEvent(_SC, _EventType, _Size, _curSize, _CacheAccess)
  If _SC#\Event 
    _SC#\Event(_SC, _EventType, _Size, _curSize, _CacheAccess)
  EndIf
EndMacro

; Wrapper fr das ECF - Callback
Procedure _SC_ECFFileAccessWrapper(ECF_File, AccessType, Size.q, curSize.q, CacheAccess, *SC.SC_Data)
  With *SC
    _SC_SendEvent(*SC, AccessType, Size.q, curSize.q, CacheAccess)
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;- Public
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;  Datei ffnen/erstellen
;  File$: Quelldatei. Falls vorhanden wird versucht sie zu ffnen, ansonst erstellt
;  Key$: Das Passwort fr das File
;  BlockSize: Blockgre der Verschlsselung, Default 256 Kb
;  MaximumCacheSize: Maximale Gre des Zwischenspeichers, Default 256 MByte
;
;  Rckgabe: SC - Handle, 0 wenn die Datei nicht erstellt werden konnte/keine SC Datei ist/beschdigt ist. 
; ------------------------------------------------------------------------------------
Procedure SC_Open(File$, Key$ = "", BlockSize.q = #SC_DefaultBlockSize, MaximumCacheSize = #SC_DefaultCacheSize)
  Protected SCHead.SC_Head, i, ii, *FreeSpaceMem
  Protected *SCE.SC_Entry, tGMem, tGMemSize.l, tGMemUnPackedSize, *tGMemBuffer, *tGMemUnpackedBuffer, isDataPacked.a
  Protected *SC.SC_Data = AllocateMemory(SizeOf(SC_Data))
  Protected *SCEXE.SC_Entry_XtraData_Entry
  Protected PackerInfo.Packer_Info
  
  With *SC
    \ECF = ECF_OpenFile(File$, Key$, BlockSize, MaximumCacheSize/2)
    If \ECF
      
      ; Wrapper Callback erstellen
      ECF_SetAccessCallback(\ECF, @_SC_ECFFileAccessWrapper(), *SC)
      
      ; Basics erstellen
      \FSC = FSC_New() ; FreeSpace Controller
      \Entrys = LL_New() ; Eintrge
      \EntryHT = HT_New() ; Eintrge Hashtable
      \Logger = Logger_New()
      \HTLimit = #SC_HT_Limit ; Ab wann wird die Hashtable verwendet?
      \UDLimit = #SC_UD_Limit ; Ab wann wird der UD Cache verwendet?
      \UseUDCache = ((MaximumCacheSize/2) Or #Null)
      If \UseUDCache
        \UDCache = Cache_New(MaximumCacheSize/2)
      EndIf
      
      ; Log schreiben
      Logger_Log(\Logger, "opening sc file")
      Logger_Log(\Logger, "total memory cache: " + Str(MaximumCacheSize) + " bytes")
      
      If ECF_FileSize(\ECF) ; Da ist schon was, laden
        ; Log schreiben
        Logger_Log(\Logger, "reading index")
        ; An Nullposition gehen
        ECF_Seek(\ECF, #Null)
        ECF_ReadData(\ECF, @SCHead, SizeOf(SC_Head)) ; Offset des Registers ermitteln
        
        ; Log schreiben
        Logger_Log(\Logger, "entry count: " + Str(SCHead\EntryCount))
        
        If SCHead\Magic <> #SC_Magic ; Header prfen
          ; Header inkorrekt, kann Datei nicht ffnen
          FSC_Free(\FSC)
          LL_Free(\Entrys)
          HT_Free(\EntryHT)
          If \UseUDCache
            Cache_Free(\UDCache)
          EndIf  
          Logger_Free(\Logger)
          ECF_CloseFile(\ECF)
          FreeMemory(*SC)
          ProcedureReturn #Null
        EndIf
        
        ; Ans Ende der Datei springen und Index lesen
        \CPos = SCHead\OffSet
        ECF_Seek(\ECF, \CPos)
        
        ; GMem anlegen, Daten aus ECF lesen
        ECF_ReadData(\ECF, @isDataPacked, SizeOf(Ascii)) ; sind die Daten gepackt?
        ECF_ReadData(\ECF, @tGMemSize, SizeOf(Long)) ; gre der Daten
        ECF_ReadData(\ECF, @tGMemUnPackedSize, SizeOf(Long)) ; ungepackte gre der Daten
        *tGMemBuffer = AllocateMemory(tGMemSize) 
        ECF_ReadData(\ECF, *tGMemBuffer, tGMemSize) ; Daten lesen
        
        ; GMem aufbauen
        tGMem = GMem_New()
        If isDataPacked ; Sind die Daten gepackt erst entpacken
          *tGMemUnpackedBuffer = Packer_Uncompress(*tGMemBuffer, tGMemSize, tGMemUnPackedSize, #SC_LZMA) ; Entpacken
          GMem_SetMem(tGMem, *tGMemUnpackedBuffer, tGMemUnPackedSize)
          FreeMemory(*tGMemBuffer)
        Else ; Einfach Speicher laden
          GMem_SetMem(tGMem, *tGMemBuffer, tGMemSize)
        EndIf
        
        ; Eintrge laden
        For i = 1 To SCHead\EntryCount
          *SCE = AllocateMemory(SizeOf(SC_Entry))
          
          GMem_ReadData(tGMem, *SCE, SizeOf(Word)) ; Nur gre des Eintrags lesen
          
          If *SCE\OwnSize <> SizeOf(SC_Entry) ; Eintrag ist in einem anderem Format gespeichert :P
            ; Kompatibilittsmodus!
            
            ; Gre ndern falls ntig
            If *SCE\OwnSize > SizeOf(SC_Entry)
              *SCE = ReAllocateMemory(*SCE, *SCE\OwnSize)
            EndIf
          EndIf
          
          ; restliche Daten lesen
          GMem_ReadData(tGMem, *SCE + SizeOf(Word), *SCE\OwnSize-(SizeOf(SC_Entry_Dynamics)+SizeOf(Word)))
          
          ; Dynmische Daten: Name
          *SCE\Dyn\Name = AllocateMemory(*SCE\NameSize)
          GMem_ReadData(tGMem, *SCE\Dyn\Name, *SCE\NameSize)
          ; Index:
          *SCE\Dyn\TId = \TIdIndex
          \TIdIndex + 1
          
          
          If *SCE\XtraDataCount ; ExtraEintrge laden
            *SCE\Dyn\XtraList = LL_New() 
            ; Alle Extra-Eintrge durchgehen
            For ii = 1 To *SCE\XtraDataCount
              *SCEXE = AllocateMemory(SizeOf(SC_Entry_XtraData_Entry))
              
              ; Allgemeine Infos + Namen laden
              GMem_ReadData(tGMem, *SCEXE, SizeOf(SC_Entry_XtraData_Entry)-SizeOf(SC_Entry_XtraData_Entry_Dynamics))
              *SCEXE\Dyn\Name = AllocateMemory(*SCEXE\NameSize)
              GMem_ReadData(tGMem, *SCEXE\Dyn\Name, *SCEXE\NameSize)
              
              ; Typspezifisches laden
              If *SCEXE\Type = #PB_String
                *SCEXE\Dyn\Value = AllocateMemory(*SCEXE\StringSize)
                GMem_ReadData(tGMem, *SCEXE\Dyn\Value, *SCEXE\StringSize)
              Else
                GMem_ReadData(tGMem, @*SCEXE\Dyn\Value, SizeOf(Quad))
              EndIf
              
              ; zu Liste hinzufgen
              LL_Add(*SCE\Dyn\XtraList) : LL_Set(*SCE\Dyn\XtraList, *SCEXE)
            Next
          EndIf
          
          ; Verfgbar machen
          LL_Add(\Entrys) : LL_Set(\Entrys, *SCE) ; zur Liste hinzufgen
          HT_SetValue(\EntryHT, PeekS(*SCE\Dyn\Name), *SCE) ; in Hashtable eintragen
          
          ; Log schreiben
          Logger_Log(\Logger, "loaded entry: " + PeekS(*SCE\Dyn\Name) + ", tId: " + Str(*SCE\Dyn\tId))
          
        Next
        
        ; FreeSpace laden
        If SCHead\FSCSize
          *FreeSpaceMem = AllocateMemory(SCHead\FSCSize)
          GMem_ReadData(tGMem, *FreeSpaceMem, SCHead\FSCSize)
          FSC_LoadFromMem(\FSC, *FreeSpaceMem, SCHead\FSCSize)
        EndIf
        
        ; GMem Freigeben
        GMem_Free(tGMem)
        
      Else
        \CPos = SizeOf(SC_Head)
      EndIf
    Else
      FreeMemory(*SC)
      *SC = #Null
    EndIf
  EndWith
  
  ProcedureReturn *SC
EndProcedure

; ------------------------------------------------------------------------------------
;  Datei schlieen
;  *SC: SC Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure SC_Close(*SC.SC_Data)
  Protected SCHead.SC_Head, i, *FreeSpaceMem, tGMem, tGMemSize.l, tGMemUnPackedSize.l
  Protected *SCE.SC_Entry, *SCEXE.SC_Entry_XtraData_Entry, *tGMem
  Protected PackerInfo.Packer_Info
  
  With *SC
    ; Datei abschlieen
    
    ; Das soll der User nicht mehr mitbekommen
    ECF_SetAccessCallback(\ECF, #Null)
    
    ; Wurde der Conatainer verndert?
    If \ContainerChanged
      
      ; Header erstellen
      SCHead\Magic = #SC_Magic
      SCHead\EntryCount = LL_Count(\Entrys)
      SCHead\OffSet = \CPos
      *FreeSpaceMem = FSC_SaveToMem(\FSC, @SCHead\FSCSize)
    
      ; Header Schreiben
      ECF_Seek(\ECF, #Null)
      ECF_WriteData(\ECF, @SCHead, SizeOf(SCHead))
    
      ; GrowableMem erstellen
      tGMem = GMem_New()
    
      ; Eintrge schreiben, in GMem zusammenfassen
      LL_Reset(\Entrys)
      While LL_Next(\Entrys)
  
        ; Indexeintrge schreiben
        *SCE = LL_Get(\Entrys)
        LL_Delete(\Entrys)
        If *SCE\Dyn\XtraList
          *SCE\XtraDataCount = LL_Count(*SCE\Dyn\XtraList)
        EndIf
      
        GMem_WriteData(tGMem, *SCE, *SCE\OwnSize-SizeOf(SC_Entry_Dynamics))
        GMem_WriteData(tGMem, *SCE\Dyn\Name, *SCE\NameSize)
      
        ; Extra - Eintrge schreiben
        If *SCE\Dyn\XtraList
          LL_Reset(*SCE\Dyn\XtraList)
          While LL_Next(*SCE\Dyn\XtraList) ; Alle ExtraDaten durchgehen
          
            *SCEXE = LL_Get(*SCE\Dyn\XtraList)
            LL_Delete(*SCE\Dyn\XtraList)
          
            ; Allgemeine Infos + Namen speichern
            GMem_WriteData(tGMem, *SCEXE, SizeOf(SC_Entry_XtraData_Entry)-SizeOf(SC_Entry_XtraData_Entry_Dynamics))
            GMem_WriteData(tGMem, *SCEXE\Dyn\Name, *SCEXE\NameSize)
            FreeMemory(*SCEXE\Dyn\Name)
            
            ; Typspezifisches Speichern
            If *SCEXE\Type = #PB_String
              GMem_WriteData(tGMem, *SCEXE\Dyn\Value, *SCEXE\StringSize)
              FreeMemory(*SCEXE\Dyn\Value)
            Else
              GMem_WriteData(tGMem, @*SCEXE\Dyn\Value, SizeOf(Quad))
            EndIf
          
            FreeMemory(*SCEXE)
          Wend
        
          LL_Free(*SCE\Dyn\XtraList)
        EndIf
           
        FreeMemory(*SCE\Dyn\Name)
        FreeMemory(*SCE)
      Wend 
    
      ; FSC schreiben
      If SCHead\FSCSize
        GMem_WriteData(tGMem, *FreeSpaceMem, SCHead\FSCSize)  
      EndIf
    
      ; GMem laden und Packen
      tGMemSize = GMem_GetMemSize(tGMem)
      tGMemUnPackedSize = tGMemSize 
      *tGMem = GMem_GetMem(tGMem)
    
      ; Packen
      PackerInfo\Packer = #SC_LZMA
      PackerInfo\CompressionLevel = 9
      PackerInfo\DictionarySize = #LZMA_DICT_4M
      *tGMem = Packer_Compress(*tGMem, tGMemSize, @PackerInfo)
    
      ; Packen berprfen
      If PackerInfo\DataOk
        tGMemSize = PackerInfo\MemSize
      Else ; Ungepackte Daten speichern
        *tGMem = GMem_GetMem(tGMem)
      EndIf
    
      ; Index real schreiben 
      ECF_Seek(\ECF, \CPos)
      ECF_WriteData(\ECF, @PackerInfo\DataOk, SizeOf(Ascii)) ; Sind die Daten richtig gepackt?
      ECF_WriteData(\ECF, @tGMemSize, SizeOf(Long)) ; Gre der Daten schreiben
      ECF_WriteData(\ECF, @tGMemUnPackedSize, SizeOf(Long)) ; Reale Gre der Daten schreiben
      ECF_WriteData(\ECF, *tGMem, tGMemSize) ; Daten schreiben
      GMem_Free(tGMem) ; GMem Freigeben
      If PackerInfo\DataOk ; gepackte Daten freigeben
        FreeMemory(*tGMem)
      EndIf
      
    EndIf
    
    ; Aufrumen
    If \Current\Mem ; Aktuellen Speicher freigeben
      FreeMemory(\Current\Mem)
    EndIf
    
    If \UseUDCache
      Cache_Free(\UDCache)  
    EndIf
    ECF_CloseFile(\ECF)
    FSC_Free(\FSC)
    HT_Free(\EntryHT)
    LL_Free(\Entrys)
    Logger_Free(\Logger)
    FreeMemory(*SC)
    
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Feineinstellungen fr die SC - Verwaltung
;  *SC: SC Handle
;  MaximumSize: Maximale Gre des Cache in Bytes, <0 = Unbegrenzt
;  AgeTolerance: AltersToleranz bei der beim Speicherfreirumen noch nach schlechteren Daten gesucht wird
;  AccessTolerance: ZugriffssToleranz bei der beim Speicherfreirumen noch nach schlechteren Daten gesucht wird
;  MinimumEntryLifeTime: Minimale Lebenszeit eines Eintrages in Sekunden
;  HTLimit: Ab wie vielen Eintrgen wird in der Hashtabelle statt in einer Liste nach den Eintrgen gesucht?
;  UDLimit: Ab wie vielen Zugriffen kann ein Eintrag im UnpackedData - Cache landen?
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure SC_Configure(*SC.SC_Data, MaximumSize = #SC_DefaultCacheSize, AgeTolerance = #Null, AccessTolerance = #Null, MinimumEntryLifeTime = 30, HTLimit = #SC_HT_Limit, UDLimit = #SC_UD_Limit)
  With *SC
    Logger_Log(\Logger, "configuring system")
    ECF_ConfigureCache(\ECF, MaximumSize/2, AgeTolerance, AccessTolerance, MinimumEntryLifeTime)
    Cache_Config(\UDCache, MaximumSize/2, AgeTolerance, AccessTolerance)
    \HTLimit = HTLimit
    \UDLimit = UDLimit
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Aktiviert das Zeitversetzte schreiben
;  *SC: SC Handle
;  Enable: #True = aktiviert, #False deaktiviert
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure SC_Transaction(*SC.SC_Data)
  With *SC
    ECF_Transaction(\ECF)
    \EnableDelayedWrite = #True
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Schreibt den Cache sicher in die ECF Datei
;  *SC: SC Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure SC_Commit(*SC.SC_Data, Enable = #False)
  With *SC
    ECF_Commit(\ECF)
    \EnableDelayedWrite = #False
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Setzt das Event-Callback
;  *SC: ECF Handle
;  *CB: Pointer auf Callback
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure SC_SetEventCallback(*SC.SC_Data, *CB.SC_EventCallback)
  With *SC
    \Event = *CB
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Datei neu Blocken
;  *SC: SC Handle
;  NewBlockSize: Neue Blockgre
;  TempFileName$: Temporrfile zum Reblocken, wenn leergelassen im Temporrverzeichnis
;
;  Rckgabe: #True wenn geReblockt werden konnte, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_ReBlock(*SC.SC_Data, NewBlockSize.q = #SC_DefaultBlockSize, TempFileName$ = "")
  With *SC
    ProcedureReturn ECF_ReBlock(\ECF, NewBlockSize, TempFileName$)  
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Entfernt ungentzten Platz aus der Datei
;  *SC: SC Handle
;
;  Rckgabe: #True wenn Platz entfernt werden konnte, #False wen nicht
; ------------------------------------------------------------------------------------
Procedure SC_Shrink(*SC.SC_Data)
  Protected Change, *SCE.SC_Entry, *SCE2.SC_Entry, *Mem
  
  With *SC
    
    If FSC_TotalFreeSpace(\FSC) ; Nur ntig wenn ungenutzter Platz da ist
      
      Logger_Log(\Logger, "shrinking file")
      
      ; Muss erst nach Position sortiert werden da sonst beim neusetzen Blcke berschrieben werden knnten
      ; Bubble Sort
      Change = #True
      While Change
        Change = #False
        LL_Reset(\Entrys)
        If LL_Next(\Entrys)
          *SCE = LL_Get(\Entrys)
          While LL_Next(\Entrys)
            *SCE2 = LL_Get(\Entrys)
            If *SCE2\Position<*SCE\Position
              LL_SwitchPrev(\Entrys)
              Change = #True
            EndIf
            *SCE = *SCE2
          Wend
        EndIf
      Wend
      
      ; Current Pos auf Anfang setzen
      \CPos = SizeOf(SC_Head)
      LL_Reset(\Entrys)
      ; Alle Elemente durchgehen
      While LL_Next(\Entrys)
        *SCE = LL_Get(\Entrys)
        
        ; Block an neue Stelle setzen
        *Mem = AllocateMemory(*SCE\PackedSize)
        ECF_Seek(\ECF, *SCE\Position)
        ECF_ReadData(\ECF, *Mem, *SCE\PackedSize)
        *SCE\Position = \CPos
        ECF_Seek(\ECF, *SCE\Position)
        ECF_WriteData(\ECF, *Mem, *SCE\PackedSize)
        FreeMemory(*Mem)
        
        ; Current Pos weiterzhlen
        \CPos + *SCE\PackedSize
      Wend
      
      ; Datei abschneiden
      ECF_Seek(\ECF, \CPos)
      ECF_Truncate(\ECF)
      
      ; FreeSpace Controller rcksetzen
      FSC_Free(\FSC)
      \FSC = FSC_New()
      
      ; Container wurde verndert
      \ContainerChanged = #True
      
      ProcedureReturn #True
    EndIf
    
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Eintrag hinzufgen
;  *SC: SC Handle
;  Name$: Name unter dem der Eintrag spter angesprochen wird
;  *Mem: Speicher der eingetragen werden soll
;  MemSize: Speichergre
;  Packer: Welcher Packe soll verwendet werden? (zLib, LZMA, bZip2, aPLib)
;  CompressionLevel: Strke der Kompression (Unwesentlich fr #SC_aPLib)
;  DictionarySize: Gre des Wrtebuches fr den Packer, nur fr #SC_LZMA von Bedeutung. (siehe lzma.pbi)
;
;  Rckgabe: #True falls Erfolg, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_AddEntry(*SC.SC_Data, Name$, *Mem, MemSize.q, Packer = #SC_zLib, CompressionLevel = 7, DictionarySize = #LZMA_DICT_4M)
  Protected *SCE.SC_Entry, FSCfound, Offset.q
  Protected PackerInfo.Packer_Info, *PackedMem, pMemSize.q
  
  With *SC
    *SCE = _SC_FindEntry(*SC, @Name$)
    If *SCE
      ; Current Setzen
      ; Log schreiben
      Logger_Log(\Logger, "entry already exist: "+Name$, #Logger_Warning)
      \Current\Entry = *SCE
      ProcedureReturn #False  
    EndIf
    
    ; Event absetzen
    _SC_SendEvent(*SC, #SC_Event_CompressStart, MemSize, MemSize, #Null)
    
    ; Packer einstellen
    PackerInfo\Packer = Packer
    PackerInfo\CompressionLevel = CompressionLevel
    PackerInfo\DictionarySize = DictionarySize
    
    ; Komprimierung der Daten
    *PackedMem = Packer_Compress(*Mem, MemSize, @PackerInfo)
    If PackerInfo\DataOk ; Kompression Ok?
      pMemSize = PackerInfo\MemSize  
    Else
      If PackerInfo\CompressionLevel <> 0
        ; Log schreiben
        Logger_Log(\Logger, "can't compress data", #Logger_Warning)  
      EndIf
      *PackedMem = *Mem
      pMemSize = MemSize
    EndIf
    
    ; Event absetzen
    _SC_SendEvent(*SC, #SC_Event_CompressFinished, MemSize, pMemSize, #Null)
   
    ; Log schreiben
    Logger_Log(\Logger, "added entry: " + Name$)
   
    ; Sucht einen passenden Platz fr die Daten
    OffSet = FSC_Find(\FSC, pMemSize, @FSCfound)
    If Not FSCfound ; Gibt Nix freies in der Datei, Eintrag muss hinten dran
      OffSet = \CPos 
    EndIf
    
    ; Listen-Eintrag erstellen
    *SCE = AllocateMemory(SizeOf(SC_Entry))
    *SCE\OwnSize = SizeOf(SC_Entry) ; Gre des Eintrages
    *SCE\Position = OffSet ; Position
    *SCE\Size = MemSize ; Orginalgre
    *SCE\PackedSize = pMemSize ; Gepackte gre
    *SCE\Compressed = PackerInfo\DataOk ; wurde korrekt gepackt?
    *SCE\NameSize = StringByteLength(Name$) + SizeOf(Character) ; Gre des Namens
    *SCE\Packer = Packer ; Welcher Packer wurde verwendet?
    *SCE\Dyn\Name = AllocateMemory(*SCE\NameSize) ; Namensspeicher anlegen
    PokeS(*SCE\Dyn\Name, Name$, *SCE\NameSize) ; Namen schreiben
    *SCE\Dyn\TId = \TIdIndex ; Index zuteilen
    \TIdIndex + 1
    
    ; Verfgbar machen
    LL_Add(\Entrys) : LL_Set(\Entrys, *SCE) ; In Liste eintragen
    HT_SetValue(\EntryHT, PeekS(*SCE\Dyn\Name), *SCE) ; In Hashtable eintragen
    
    ECF_Seek(\ECF, Offset)
    ECF_WriteData(\ECF, *PackedMem, pMemSize)
    
    If Not FSCfound ; Zeiger weiterstellen wenn ntig
      \CPos + pMemSize
    EndIf
    
    If PackerInfo\DataOk ; Speicher freigeben
      FreeMemory(*PackedMem)
    EndIf
    
    ; Current Setzen
    \Current\Entry = *SCE
    
    ; Container wurde verndert
    \ContainerChanged = #True
    
    ProcedureReturn #True
  EndWith
  
EndProcedure

; ------------------------------------------------------------------------------------
;  Eintrag laden
;  *SC: SC Handle
;  Name$: Name des Eintrags
;  LoadMem: Wenn True wird der Eintrag in den Speicher geladen, sonst nicht. Ntzlich wenn man nur alle Eintrge auflisten mchte.
;
;  Rckgabe: #True wenn gefunden, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_GetEntry(*SC.SC_Data, Name$, LoadMem = #True)
  Protected RtVar = #False
  Protected *SCE.SC_Entry
  
  With *SC
    
    ; Erkennt Zugriffe auf immer dasselbe Element
    If \Current\Entry And (CompareMemoryString(\Current\Entry\Dyn\Name, @Name$) = #PB_String_Equal)
      \Current\Entry\Dyn\TAccessCount + 1 ; Zugriffszhler erhhen
      If LoadMem
        If \Current\Mem
          ProcedureReturn #True
        EndIf
      Else
        ProcedureReturn #True
      EndIf
    EndIf
    
    *SCE = _SC_FindEntry(*SC, @Name$) ; Eintrag suchen
    If *SCE ; Es gibt den Eintrag :D
      If \Current\Mem ; Aktuellen Speicher freigeben
        FreeMemory(\Current\Mem)
      EndIf
      \Current\Entry = *SCE ; EIntrag aktuell setzen
      
      *SCE\Dyn\TAccessCount + 1 ; Zugriffszhler erhhen
      
      If LoadMem ; soll auch der Speicher geladen werden?
        _SC_LoadEntry(*SC)
        ; Und Fertig :D
        If \Current\Mem
          RtVar = #True
        EndIf
      Else
        \Current\Mem = #Null
        RtVar = #True
      EndIf
      
      ; Log schreiben
      Logger_Log(\Logger, "loaded entry: " + Name$)
    
    Else
      
      ; Log schreiben
      Logger_Log(\Logger, "can't load entry: " + Name$)
      
    EndIf
  EndWith
  
  ProcedureReturn RtVar
EndProcedure

; ------------------------------------------------------------------------------------
;  Eintrag lschen
;  *SC: SC Handle
;  Name$: Name des Eintrags
;  SafeDelete: Der Platz an dem der Eintrag bisher lag wird mit Zufallsdaten berschrieben
;
;  Rckgabe: #True wenn der Eintrag gelscht werden konnte, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_Delete(*SC.SC_Data, Name$, SafeDelete = #False)
  Protected RtVar = #False
  Protected *SCE.SC_Entry, *SCEXE.SC_Entry_XtraData_Entry
  
  With *SC
    *SCE = _SC_FindEntry(*SC, @Name$, #False)
    
    If *SCE
      FSC_Add(\FSC, *SCE\Position, *SCE\PackedSize) ; Freigegebenen Speicherplatz in den Controller speichern
      
      ; Eintrag aus Hashtable lschen
      HT_SetValue(\EntryHT, PeekS(*SCE\Dyn\Name), #Null)
      ; Eintrag lschen
      LL_Delete(\Entrys)
      
      ; Name freigeben
      FreeMemory(*SCE\Dyn\Name)
      
      ; Xtra-Eintrge lschen wenn da
      If *SCE\Dyn\XtraList
        LL_Reset(*SCE\Dyn\XtraList)
        While LL_Next(*SCE\Dyn\XtraList) ; Alle ExtraDaten durchgehen
          *SCEXE = LL_Get(*SCE\Dyn\XtraList)
          LL_Delete(*SCE\Dyn\XtraList)
          FreeMemory(*SCEXE\Dyn\Name)
          If *SCEXE\Type = #PB_String
            FreeMemory(*SCEXE\Dyn\Value)
          EndIf
          FreeMemory(*SCEXE)
        Wend
        LL_Free(*SCE\Dyn\XtraList)
      EndIf
      
      If SafeDelete ; Sicheres lschen (berschreiben mit Zufallswerten)
        ECF_Seek(\ECF, *SCE\Position)
        ECF_SafeDelete(\ECF, *SCE\PackedSize)
      EndIf
      
      ; Wenn der Delayed Write aktiv ist nicht lschen, sonst kontraproduktiv
      If Not \EnableDelayedWrite
        ECF_Seek(\ECF, *SCE\Position)
        ECF_ClearCache(\ECF, *SCE\PackedSize)
      EndIf
      
      If *SC\UseUDCache ; Aus dem Unpacked Cache lschen
        Cache_DeleteEntry(*SC\UDCache, *SCE\Dyn\tId)
      EndIf
      
      FreeMemory(*SCE)
      
      ; Log schreiben
      Logger_Log(\Logger, "deleted entry: " + Name$)
      
      ; Container wurde verndert
      \ContainerChanged = #True
      
      ; Und Fertig :D
      RtVar = #True

    Else
      
      ; Log schreiben
      Logger_Log(\Logger, "can't delete entry: " + Name$)
      
    EndIf
  EndWith
  
  ProcedureReturn RtVar
EndProcedure

; ------------------------------------------------------------------------------------
;  Enumeration starten
;  *SC: SC Handle
;  LoadMem: Wenn True wird der aktuelle Eintrag in den Speicher geladen, sonst nicht.
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure SC_EnumerateEntrys(*SC.SC_Data, LoadMem = #True)
  With *SC
    \EnumLoadMem = LoadMem
    LL_Reset(\Entrys)
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Nchste Datei in der Enumeration aufrufen
;  *SC: SC Handle
;
;  Rckgabe: #True wenn vorhanden, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_NextEntry(*SC.SC_Data)
  Protected isNext
  Protected *SCE.SC_Entry
  
  With *SC
    isNext = LL_Next(\Entrys)
    If isNext
      If \Current\Mem ; Aktuellen Speicher freigeben
        FreeMemory(\Current\Mem)
      EndIf
      *SCE = LL_Get(\Entrys)
      \Current\Entry = *SCE
      
      *SCE\Dyn\TAccessCount + 1 ; Zugriffszhler erhhen
      
      If \EnumLoadMem ; Speicher laden
        _SC_LoadEntry(*SC)
      Else
        \Current\Mem = #Null
      EndIf
      
    EndIf
    ProcedureReturn isNext  
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Name des aktuellen Eintrags
;  *SC: SC Handle
;
;  Rckgabe: Name des Eintrags
; ------------------------------------------------------------------------------------
Procedure$ SC_GetName(*SC.SC_Data)
  With *SC
    If \Current\Entry
      ProcedureReturn PeekS(\Current\Entry\Dyn\Name)
    EndIf
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Name des aktuellen Eintrags ndern
;  *SC: SC Handle
;
;  Rckgabe: #True falls Erfolg, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_SetName(*SC.SC_Data, Name$)
  With *SC
    If \Current\Entry
      
      ; Log schreiben
      Logger_Log(\Logger, "changed name: " + PeekS(\Current\Entry\Dyn\Name) + " -> " + Name$)
      
      ; Alten Namen freigeben
      FreeMemory(\Current\Entry\Dyn\Name)
      ; Neuen Namen schreiben
      \Current\Entry\NameSize = StringByteLength(Name$)+SizeOf(Character)
      \Current\Entry\Dyn\Name = AllocateMemory(\Current\Entry\NameSize)
      PokeS(\Current\Entry\Dyn\Name, Name$)
      
      ; Container wurde verndert
      \ContainerChanged = #True
      
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Speicher des aktuellen Eintrags
;  *SC: SC Handle
;
;  Rckgabe: Speicher des Eintrags
; ------------------------------------------------------------------------------------
Procedure SC_GetMem(*SC.SC_Data)
  With *SC
    ProcedureReturn \Current\Mem  
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Speichergre des aktuellen Eintrags
;  *SC: SC Handle
;
;  Rckgabe: Speichergre des Eintrags
; ------------------------------------------------------------------------------------
Procedure.q SC_GetMemSize(*SC.SC_Data)
  With *SC
    If \Current\Entry
      ProcedureReturn \Current\Entry\Size
    EndIf
  EndWith
EndProcedure


; ------------------------------------------------------------------------------------
;  gre des gepackten Eintrags in der Datei
;  *SC: SC Handle
;
;  Rckgabe:Gre des Eintrags
; ------------------------------------------------------------------------------------
Procedure.q SC_GetCompressedSize(*SC.SC_Data)
  With *SC
    If \Current\Entry
      ProcedureReturn \Current\Entry\PackedSize
    EndIf
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Fgt eine Extra - StringInformation zum aktuellen Eintrag hinzu
;  *SC: SC Handle
;  Name$: Name der Information
;  Value$: ExtraInformation
;
;  Rckgabe: #True wenn erfolg, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_SetXtraDataString(*SC.SC_Data, Name$, Value$)
  _SC_SetXtraData(*SC, Name$, #PB_String, @Value$)
EndProcedure

; ------------------------------------------------------------------------------------
;  Fgt eine Extra - QuadInformation zum aktuellen Eintrag hinzu
;  *SC: SC Handle
;  Name$: Name der Information
;  Value: ExtraInformation
;
;  Rckgabe: #True wenn erfolg, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_SetXtraDataQuad(*SC.SC_Data, Name$, Value.q)
  _SC_SetXtraData(*SC, Name$, #PB_Quad, Value)
EndProcedure

; ------------------------------------------------------------------------------------
;  Ld eine StringInformation vom aktuellen Eintrag
;  *SC: SC Handle
;  Name$: Name der Information
;
;  Rckgabe: Information
; ------------------------------------------------------------------------------------
Procedure.s SC_GetXtraDataString(*SC.SC_Data, Name$)
  Protected *SCEXE.SC_Entry_XtraData_Entry
  
  *SCEXE = _SC_FindXtraData(*SC, @Name$)
  If *SCEXE
    If *SCEXE\Type = #PB_String And *SCEXE\Dyn\Value; Typenprfung
      ProcedureReturn PeekS(*SCEXE\Dyn\Value)
    EndIf
  EndIf
EndProcedure

; ------------------------------------------------------------------------------------
;  Ld eine QuadInformation vom aktuellen Eintrag
;  *SC: SC Handle
;  Name$: Name der Information
;
;  Rckgabe: Information
; ------------------------------------------------------------------------------------
Procedure.q SC_GetXtraDataQuad(*SC, Name$)
  Protected *SCEXE.SC_Entry_XtraData_Entry
  
  *SCEXE = _SC_FindXtraData(*SC, @Name$)
  If *SCEXE
    If *SCEXE\Type <> #PB_String ; Typenprfung
      ProcedureReturn *SCEXE\Dyn\Value
    EndIf
  EndIf
EndProcedure

; ------------------------------------------------------------------------------------
;  Prft ob ein Extra-Eintrag existiert
;  *SC: SC Handle
;  Name$: Name der Information
;
;  Rckgabe: #True wenn der Eintrag Existiert, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_XtraDataExists(*SC.SC_Data, Name$)
  
  ProcedureReturn (_SC_FindXtraData(*SC, @Name$) Or #Null)
EndProcedure

; ------------------------------------------------------------------------------------
;  Lscht einen Extra-Eintrag
;  *SC: SC Handle
;  Name$: Name der Information
;
;  Rckgabe: #True wenn der Eintrag gelscht werden konnte, sonst #False
; ------------------------------------------------------------------------------------
Procedure SC_DeleteXtraData(*SC.SC_Data, Name$)
  Protected *SCEXE.SC_Entry_XtraData_Entry
  
  With *SC
    
    ; Eintrag suchen
    *SCEXE = _SC_FindXtraData(*SC, @Name$)
    If *SCEXE ; Eintrag ist vorhanden
      
      LL_Delete(\Current\Entry\Dyn\XtraList) ; Eintrag lschen
      If LL_Count(\Current\Entry\Dyn\XtraList) = #Null ; Liste lschen wenn sie leer ist.
        LL_Free(\Current\Entry\Dyn\XtraList)
        \Current\Entry\Dyn\XtraList = #Null  ; Liste Nullsetzen
      EndIf
    
      FreeMemory(*SCEXE\Dyn\Name) ; Namenspeicher freigeben
    
      If *SCEXE\Type = #PB_String ; Eintrag ist ein String => Stringspeicher freigeben
        FreeMemory(*SCEXE\Dyn\Value)
      EndIf
    
      FreeMemory(*SCEXE) ; Eintrag freigeben
      
      ; Container wurde verndert
      \ContainerChanged = #True
      
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure


Procedure SC_LoggerEnable(*SC.SC_Data, Enable = #True)
  
  With *SC
    Logger_Enable(\Logger, Enable)
  EndWith    
EndProcedure

Procedure SC_LoggerClear(*SC.SC_Data)
  
  With *SC
    Logger_Clear(\Logger)  
  EndWith
EndProcedure

Procedure SC_LoggerStartView(*SC.SC_Data, MinimumPriority = #Logger_Info)
  
  With *SC
    Logger_StartView(\Logger, MinimumPriority)  
  EndWith
EndProcedure

Procedure SC_LoggerViewNext(*SC.SC_Data)
  
  With *SC
    ProcedureReturn Logger_ViewNext(\Logger)  
  EndWith
EndProcedure


Procedure SC_LoggerGetDate(*SC.SC_Data)
  
  With *SC
    ProcedureReturn SC_LoggerGetDate(\Logger)  
  EndWith
EndProcedure

Procedure SC_LoggerGetPriority(*SC.SC_Data)
  
  With *SC
    ProcedureReturn SC_LoggerGetPriority(\Logger)  
  EndWith
EndProcedure

Procedure.s SC_LoggerGetMessage(*SC.SC_Data)
  
  With *SC
    ProcedureReturn SC_LoggerGetMessage(\Logger)  
  EndWith
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 24
; Folding = -------
; EnableXP