﻿ ; EnableExplicit

; bzip2 1.05
; http://www.bzip.org/

; Copyright (C) 1996-2008 Julian Seward <jseward@bzip.org>
; bzip2 is a freely available, patent free, high-quality data compressor. It 
; It typically compresses files to within 10% To 15% of the best available techniques 
; (the PPM family of statistical compressors), whilst being around twice as fast at 
; compression and six times faster at decompression.

; Wrapper 1.0 for PB 4.41 by Luis (Feb 2010)

; In brief:
; compression: very good ratio, good speed, average memory requirements (from 1200 KB to 7600 KB)
; decompression: good speed, average memory requirements (same as compression)

; all ok
#BZIP2_OK               = 0

; error codes for aPLib
#BZIP2_ERROR            = 1
#BZIP2_ERROR_MEM        = 2
#BZIP2_ERROR_OUTPUT     = 3
#BZIP2_ERROR_PARAM      = 4
#BZIP2_ERROR_DATA       = 5 

CompilerIf (#PB_Compiler_Processor = #PB_Processor_x86)
 Import "libs\" + "bzip2_32.lib"   
    BZ2_bzBuffToBuffCompress.i (*DestBuff, *DestSize, *SourceBuff, iSourceSize, iBlockSize, verbosity, workFactor)    
    BZ2_bzBuffToBuffDecompress.i (*DestBuff, *DestSize, *SourceBuff, iSourceSize, small, verbosity)
 EndImport
CompilerElse   
 Import "libs\" + "bzip2_64.lib"     
    BZ2_bzBuffToBuffCompress.i (*DestBuff, *DestSize, *SourceBuff, iSourceSize, iBlockSize, verbosity, workFactor)    
    BZ2_bzBuffToBuffDecompress.i (*DestBuff, *DestSize, *SourceBuff, iSourceSize, small, verbosity) 
 EndImport
CompilerEndIf


Procedure.i CalcDestBufferSize_BZIP2 (iSourceBuffSize)
 ProcedureReturn (iSourceBuffSize + iSourceBuffSize / 100 + 600)
EndProcedure

 ; *SourceBuff  = ptr to the source buffer
 ; *DestBuff    = ptr to the destination buffer
 ; iSourceSize  = size of the source buffer
 ; *iDestSize   = ptr to the size of the destination buffer (it will contain the packed size on exit)

 ; NOTE: iBlockSize must be between 1 (lowest) and 9 (highest) - default 9 (1 = 100K [1,2MB used], 9 = 900K [7,6MB used])
 
 ; NOTE: The size of the destination buffer *should* be at least the size returned by CalcDestBufferSize_BZIP2()

 ; NOTE: The destination buffer will contain a LONG followed by the packed data
 
 ; NOTE: #BZIP2_ERROR_OUTPUT means the file cannot be compressed (or the destination buffer is too small)
Procedure.i PackMemory_BZIP2 (*SourceBuff, *DestBuff, iSourceSize, *iDestSize.Integer, iBlockSize = 9)
 Protected iResult
   
 iResult = BZ2_bzBuffToBuffCompress (*DestBuff, *iDestSize, *SourceBuff, iSourceSize, iBlockSize, 0, 30)
  
 Select iResult
    Case #BZIP2_OK 
      If *iDestSize\i >= iSourceSize
            iResult = #BZIP2_ERROR_OUTPUT
            ; return -1 for the size of the compressed data 
            PokeI(*iDestSize, -1) ; cannot compress this
        EndIf
    Case -2 ; BZ_PARAM_ERROR
        iResult = #BZIP2_ERROR_PARAM
    Case -3 ; BZ_MEM_ERROR         
        iResult = #BZIP2_ERROR_MEM
    Default
        iResult = #BZIP2_ERROR
 EndSelect   
    
 ProcedureReturn iResult
EndProcedure
  
 ; *SourceBuff  = ptr to the source buffer
 ; *DestBuff    = ptr to the destination buffer
 ; iSourceSize  = size of the source buffer
 ; *iDestSize   = ptr to the size of the destination buffer (it will contain the unpacked size on exit)
  
 ; NOTE: The source buffer must contain a LONG followed by the packed data
 
Procedure.i UnpackMemory_BZIP2 (*SourceBuff, *DestBuff, iSourceSize, *iDestSize.Integer)
 Protected iResult
 
 iResult = BZ2_bzBuffToBuffDecompress (*DestBuff, *IDestSize, *SourceBuff, iSourceSize, 0, 0)

 Select iResult
    Case #BZIP2_OK        
        ; Alles Ok
    Case -2 ; BZ_PARAM_ERROR
        iResult = #BZIP2_ERROR_PARAM
    Case -4, -5, -7 ; BZ_DATA_ERROR, BZ_DATA_ERROR_MAGIC, BZ_UNEXPECTED_EOF
        iResult = #BZIP2_ERROR_DATA
    Case -3, -8 ; BZ_MEM_ERROR, BZ_OUTBUFF_FULL,
        iResult = #BZIP2_ERROR_MEM
    Default
        iResult = #BZIP2_ERROR         
 EndSelect   
  
 ProcedureReturn iResult
EndProcedure


; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 33
; Folding = -
; EnableXP
; EnableUser
; Executable = test.exe
; CPU = 1