; ------------------------------------------------------------------------------------
; Universal Packer Include
; Based on Packer Includes from luis:
; http://www.purebasic.fr/english/viewtopic.php?f=12&t=41217
; Author: Alexander Aigner
; PB 4.4+ x86/x64 Ascii/Unicode
; ------------------------------------------------------------------------------------


XIncludeFile "zlib.pbi"
XIncludeFile "lzma.pbi"
XIncludeFile "bzip2.pbi"
XIncludeFile "aPLib.pbi"

; EnableExplicit

Enumeration
  #Packer_zLib
  #Packer_LZMA
  #Packer_bZip2
  #Packer_aPLib
  #Packer_Uncompressed = -1
EndEnumeration

Structure Packer_Info
  CompressionLevel.i
  Packer.a
  DictionarySize.i
  DataOk.a
  MemSize.i
EndStructure

Procedure Packer_Compress(*Mem, MemSize, *Info.Packer_Info)
  Protected *PackedMem, PackedMemSize
  
  Select *Info\Packer
    Case #Packer_zLib
      PackedMemSize = CalcDestBufferSize_zLib(MemSize)
      *PackedMem = AllocateMemory(PackedMemSize)
      *Info\DataOk = (Not PackMemory_zLib(*Mem, *PackedMem, MemSize, @PackedMemSize, *Info\CompressionLevel))
      
    Case #Packer_LZMA
      PackedMemSize = CalcDestBufferSize_LZMA(MemSize)
      *PackedMem = AllocateMemory(PackedMemSize)
      *Info\DataOk = (Not PackMemory_LZMA(*Mem, *PackedMem, MemSize, @PackedMemSize, *Info\CompressionLevel, *Info\DictionarySize))
      
    Case #Packer_bZip2
      PackedMemSize = CalcDestBufferSize_bZip2(MemSize)
      *PackedMem = AllocateMemory(PackedMemSize)
      *Info\DataOk = (Not PackMemory_bZip2(*Mem, *PackedMem, MemSize, @PackedMemSize, *Info\CompressionLevel))
      
    Case #Packer_aPLib
      PackedMemSize = CalcDestBufferSize_aPLib(MemSize)
      *PackedMem = AllocateMemory(PackedMemSize)
      *Info\DataOk = (Not PackMemory_aPLib(*Mem, *PackedMem, MemSize, @PackedMemSize))
      
    Case #Packer_Uncompressed
      *Info\DataOk = #False
      
  EndSelect
  
  If *Info\DataOk
    *PackedMem = ReAllocateMemory(*PackedMem, PackedMemSize)
    *Info\MemSize = PackedMemSize
  Else
    If *PackedMem
      FreeMemory(*PackedMem)
    EndIf
    *PackedMem = #Null
  EndIf
  
  ProcedureReturn *PackedMem
EndProcedure

Procedure Packer_Uncompress(*Mem, PackedSize, UnpackedMemSize, Packer)
  Protected *UnPackedMem = AllocateMemory(UnpackedMemSize)
  
  Select Packer
    Case #Packer_zLib
      If UnpackMemory_zLib(*Mem, *UnPackedMem, PackedSize, @UnpackedMemSize) <> #ZLIB_OK
        FreeMemory(*UnPackedMem)
        *UnPackedMem = #Null
      EndIf
      
    Case #Packer_LZMA
      If UnpackMemory_LZMA(*Mem, *UnPackedMem, PackedSize, @UnpackedMemSize) <> #LZMA_OK
        FreeMemory(*UnPackedMem)
        *UnPackedMem = #Null
      EndIf
      
    Case #Packer_bZip2
      If UnpackMemory_bZip2(*Mem, *UnPackedMem, PackedSize, @UnpackedMemSize) <> #BZIP2_OK
        FreeMemory(*UnPackedMem)
        *UnPackedMem = #Null
      EndIf
      
    Case #Packer_aPLib
      If UnpackMemory_aPLib(*Mem, *UnPackedMem, PackedSize, @UnpackedMemSize) <> #APLIB_OK
        FreeMemory(*UnPackedMem)
        *UnPackedMem = #Null
      EndIf
      
  EndSelect
  
  ProcedureReturn *UnPackedMem
EndProcedure

; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 12
; FirstLine = 7
; Folding = -
; EnableXP