﻿; EnableExplicit

; zlib 1.2.3
; http://www.zlib.net/

; This product uses the zlib compression library.

; Wrapper 1.0 for PB 4.41 by Luis (Feb 2010)

; In brief:
; compression: good ratio, very good speed, very low memory requirements (around 256 KB)
; decompression: extremely fast, very low memory requirements (should be less than 48 KB)

; all ok
#ZLIB_OK = 0

; error codes for zlib
#ZLIB_ERROR         = 1
#ZLIB_ERROR_MEM     = 2
#ZLIB_ERROR_OUTPUT  = 3
#ZLIB_ERROR_PARAM   = 4
#ZLIB_ERROR_DATA    = 5 

CompilerIf (#PB_Compiler_Processor = #PB_Processor_x86)
 ImportC "libs\" + "zlib_32.lib" 
    compress2(*DestBuff, *iDestSize, *SourceBuff, iSourceSize, iCompressionLevel)
    uncompress(*DestBuff, *iDestSize, *SourceBuff, iSourceSize)
    compressBound(SourceLen)
 EndImport
CompilerElse   
 ImportC "libs\" + "zlib_64.lib" 
    compress2(*DestBuff, *iDestSize, *SourceBuff, iSourceSize, iCompressionLevel)
    uncompress(*DestBuff, *iDestSize, *SourceBuff, iSourceSize)
    compressBound(SourceLen)
 EndImport
CompilerEndIf


Procedure.i CalcDestBufferSize_ZLIB (iSourceBuffSize)
 ProcedureReturn compressBound(iSourceBuffSize)
EndProcedure

 ; *SourceBuff  = ptr to the source buffer
 ; *DestBuff    = ptr to the destination buffer
 ; iSourceSize  = size of the source buffer
 ; *iDestSize   = ptr to the size of the destination buffer (it will contain the packed size on exit)
 ; NOTE: The size of the destination buffer *should* be at least the size returned by CalcDestBufferSize_ZLIB()
 ; NOTE: #ZLIB_ERROR_OUTPUT means the file cannot be compressed (or the destination buffer is too small)
 ; NOTE: iCompressionLevel must be between 0 (lowest) and 9 (highest) - default 9
Procedure.i PackMemory_ZLIB (*SourceBuff, *DestBuff, iSourceSize, *iDestSize.Integer, iCompressionLevel = 9)
 Protected iResult
 
 iResult = compress2 (*DestBuff, *iDestSize, *SourceBuff, iSourceSize, iCompressionLevel)
 
 Select iResult
    Case #ZLIB_OK
        If *iDestSize\i >= iSourceSize
          iResult = #ZLIB_ERROR_OUTPUT
            ; return -1 for the size of the compressed data 
            PokeI(*iDestSize, -1) ; cannot compress this 
        EndIf
    Case -3 ; Z_DATA_ERROR
        iResult = #ZLIB_ERROR_DATA
    Case -4, -5 ; Z_ERROR_MEM, Z_BUF_ERROR
        iResult = #ZLIB_ERROR_MEM
    Default ; ANYTHING ELSE
        iResult = #ZLIB_ERROR
 EndSelect
 
 ProcedureReturn iResult
EndProcedure
  
 ; *SourceBuff  = ptr to the source buffer
 ; *DestBuff    = ptr to the destination buffer
 ; iSourceSize  = size of the source buffer
 ; *iDestSize   = ptr to the size of the destination buffer (it will contain the unpacked size on exit)
Procedure.i UnpackMemory_ZLIB (*SourceBuff, *DestBuff, iSourceSize, *iDestSize.Integer)
  Protected iResult
   
 iResult = uncompress (*DestBuff, *iDestSize, *SourceBuff, iSourceSize)

 Select iResult
    Case #ZLIB_OK
        ; Alles Ok
    Case -3 ; Z_DATA_ERROR
        iResult = #ZLIB_ERROR_DATA
    Case -4, -5 ; Z_ERROR_MEM, Z_BUF_ERROR
        iResult = #ZLIB_ERROR_MEM        
    Default ; ANYTHING ELSE
        iResult = #ZLIB_ERROR
 EndSelect
  
 ProcedureReturn iResult
EndProcedure


; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 30
; FirstLine = 10
; Folding = -
; EnableXP
; EnableUser
; Executable = test.exe
; CPU = 1